const GlobalExceptionHandler = (function() {

    const scriptMessageHandlerName = "messageDetailScriptHandler"; // MessageRead
    const postActionName = "scriptRuntimeException";
    let webViewScriptId; // used to identify logs coming from the same read message script (lazy init)

    const Level = {
        error: "error",
        warning: "warning",
        info: "info"
    };

    const Type = {
        jsUncaughtException: "jsUncaughtException",
        resourceLoadError: "resourceLoadError",
        unknownError: "unknownError"
    };

    // Using global "error" event listener instead of window.onerror handler, because can capture both
    // script exceptions and image load errors. window.onerror receives only script exceptions.
    window.addEventListener("error", (event) => {

        // JS uncaught exception
        if (event.target == window) {
            const message = (event => {
                if (!hasValue(event.error)) {
                    return "js uncaught window exception (no event.error object)";
                }
                if (!hasValue(event.error.message)) {
                    return "js uncaught window exception (no error.message)"
                }
                return event.error.message;
            })(event);

            postMessage({
                level: Level.error,
                type: Type.jsUncaughtException,
                timing: JSON.stringify(getPageTiming()),
                state: JSON.stringify(getPageState()),
                message: message,
                stack: hasValue(event.error) && event.error.stack
            });
            return;
        }

        // Resource load error. Unfortunately there is no way to get the actual error message.
        // Also, we get no error event if a background image load fails, unlike <img> tags.
        // Only failed SCRIPT or IFRAME load is considered an error (used in AMP messages). Other elements
        // that can have load error (IMG, SOURCE, AUDIO, VIDEO, EMBED, INPUT type=image, TRACK) are
        // considered warnings and not logged in Splunk (only errors are logged in Splunk).
        if (event.target.tagName && event.target.hasAttribute('src')) {
            const isErrorLevel = /SCRIPT|IFRAME/.test(event.target.tagName);
            postMessage({
                level: isErrorLevel ? Level.error : Level.warning,
                type: Type.resourceLoadError,
                attributes: getElementAttributes(event.target),
                message: `${event.target.tagName} load failed`
            });
            return;
        }

        // Should never get here, this is unknown error case. Only SCRIPT tag in addition to IMG
        // can have load error event, but message body cannot have external scripts.
        postMessage({
            level: Level.error,
            type: Type.unknownError,
            timing: JSON.stringify(getPageTiming()),
            state: JSON.stringify(getPageState()),
            attributes: getElementAttributes(event.target),
            message: (hasValue(event.error) && event.error.message) || `Unknown error event from <${event.target.tagName || event.target}> element`,
            stack: hasValue(event.error) && event.error.stack
        });

    }, /*capturingPhase:*/true);

    function postMessage(data) {
        webViewScriptId = webViewScriptId || String(Math.round(Date.now() * (Math.random() || 1)));
        data.webViewScriptId = webViewScriptId;
        data.action = postActionName;
        window.webkit.messageHandlers[scriptMessageHandlerName].postMessage(data);
    }

    function log(message, level = Level.info) {
        postMessage({
            level: level,
            type: level,
            message: message
        });
    }

    function isNumber(value) {
        return typeof value === "number" && isFinite(value);
    }
    function hasValue(value) {
        return !(value === null || value === undefined);
    }

    function getElementAttributes(element) {
        try {
            return _getElementAttributes(element);
        } catch {
            return "";
        }
    }

    function getPageTiming() {
        try {
            return _getPageTiming();
        } catch {
            return {};
        }
    }

    function getPageState() {
        try {
            return _getPageState();
        } catch {
            return {};
        }
    }

    function _getElementAttributes(element) {
        if (!hasValue(element)) return '';

        const skipAttributesRegex = /^id|class|style$/i;
        const MAX_VALUE_LENGTH = 160;
        const ELLIPSIS_UNICODE = '\u2026';
        let results = [];

        for (const attribute of element.attributes) {
            if (skipAttributesRegex.test(attribute.name)) continue;
            if (attribute.value.trim().length == 0) continue;

            let value = attribute.value;
            if (value.length > MAX_VALUE_LENGTH) {
                value = `${value.substr(0, MAX_VALUE_LENGTH)}${ELLIPSIS_UNICODE}`;
            }
            results.push(`${attribute.name}=${value}`);
        }

        return results.join(',');
    }

    function _getPageTiming() {
        const pageTiming = {};

        const upTime = getUpTime();
        if (upTime) {
            pageTiming.upTime = upTime;
        }

        const firstPaintMark = getFirstPaintTime();
        if (firstPaintMark) {
            pageTiming.firstPaintMark = firstPaintMark;
        }

        const navigationTiming = getNavigationTiming();
        Object.assign(pageTiming, navigationTiming);

        return pageTiming;
    }

    // Use performance.now() to get how mach time has expired since the page was created
    function getUpTime() {
        const upTime = hasValue(window.performance) && typeof performance.now == "function" && performance.now();
        if (isNumber(upTime)) {
            return upTime + "ms";
        }
        return undefined;
    }

    // Use "first-contentful-paint" performance event to check if anything was rendered on the screen yet
    function getFirstPaintTime() {
        if ( !(hasValue(window.performance) && typeof performance.getEntriesByName === "function")) {
            return undefined;
        }

        const entries = performance.getEntriesByName("first-contentful-paint");
        if (entries && isNumber(entries.length) && entries.length > 0) {
            let firstPaintTime = entries[0].startTime;
            return isNumber(firstPaintTime) ? firstPaintTime + "ms" : undefined;
        }
        return undefined;
    }

    // Extract meaningful timings from the performance.timing object
    function getNavigationTiming() {
        const timeSince = (since, value) => {
            if (isNumber(value) && value > since) {
                return value - since;
            }
            return undefined;
        }

        const navigationTiming = {};

        if (!(hasValue(window.performance) && hasValue(performance.timing))) {
            return navigationTiming;
        }

        const timing = performance.timing;
        const navigationStart = timing.navigationStart;
        if (!(isNumber(navigationStart) && navigationStart > 0)) {
            return navigationTiming;
        }

        // domLoading, domInteractive and domComplete correspond to when the document’s readyState is
        // set to the corresponding loading, interactive and complete states.
        // unloadEventStart tell us if the WebView had previous content (reuse WebView), which is replaced with this message.
        for (let property of ["unloadEventStart", "domLoading", "domInteractive", "domComplete"]) {
            const value = timing[property];
            let time = timeSince(navigationStart, value);
            if (isNumber(time)) {
                let timingName = (property + "Mark").replace("Start", "");
                navigationTiming[timingName] = time + "ms";
            }
        }

        // How much time was spent in DOMContentLoaded event processing, corresponds to page setup time
        const setupStart = timeSince(navigationStart, timing.domContentLoadedEventStart);
        const setupEnd = timeSince(navigationStart, timing.domContentLoadedEventEnd);
        if (isNumber(setupStart) && isNumber(setupEnd)) {
            navigationTiming.pageSetupMeasure = (setupEnd - setupStart) + "ms";
        }

        return navigationTiming;
    }

    // Get DOM state: readyState, element count, content size
    function _getPageState() {
        const document = window.document;
        const documentElement = hasValue(document) && document.documentElement;

        const pageState = {};

        const readyState = hasValue(document) && document.readyState;
        if (readyState) {
            pageState.documentReadyState = readyState;
        }

        if (hasValue(documentElement)) {
            pageState.contentSize = `${documentElement.offsetWidth}x${documentElement.offsetHeight}`;
        }

        const domElementCount = hasValue(document) && hasValue(document.all) && document.all.length;
        if (isNumber(domElementCount)) {
            pageState.domElementCount = domElementCount;
        }

        const imageCount = hasValue(document) && hasValue(document.images) && document.images.length;
        if (isNumber(imageCount)) {
            pageState.imageCount = imageCount;
        }

        return pageState;
    }

    function getPageStatus() {
        return {
            timing: getPageTiming(),
            state: getPageState(),
        };
    }

    function logPageStatus(tag = "status") {
        log(`On ${tag} - timing: ${JSON.stringify(getPageTiming())} state: ${JSON.stringify(getPageState())}`);
    }

    // Public
    return {
        level: Level,
        log: log,
        getPageStatus: getPageStatus,
        logPageStatus: logPageStatus,
    };
}());

