var YMMessageCompose = (function() {

    /* CONSTANTS */

    let COMPOSE_AREA_ID = "yahoo__compose_area",
        FOCUS_COMPOSE_DELAY = 100,
        SCROLL_DEBOUNCE_TIME = 100,
        CURSOR_PLACEHOLDER_CLASS_NAME = "cursor-placeholder",
        DND_ACTIVE_CLASS_NAME = "dnd-active",
        CURSOR_PADDING = 5,

        /* PRIVATE PROPERTIES */

        _composeArea,
        _scrollIfNeededTimeout,

        /* PRIVATE METHODS */

        _initElements = function() {
            _composeArea = document.getElementById(COMPOSE_AREA_ID);
        },

        _postNativeMsg = function(msg, data) {
            window.webkit.messageHandlers.messageComposeScriptHandler.postMessage(JSON.stringify({
                message: msg,
                data: data || null
            }));
        },

        _removeCursorPlaceholder = function() {
            document.querySelectorAll("." + CURSOR_PLACEHOLDER_CLASS_NAME).forEach(function(el) {
                el.remove();
            });
        },

        _getPlaceholderSpan = function() {
            let span = document.createElement("span");
            span.className = CURSOR_PLACEHOLDER_CLASS_NAME;
            span.innerHTML = "&nbsp;";
            return span;
        },

        _isComposeAreaInFocus = function() {
            return document.activeElement && document.activeElement.closest('#' + COMPOSE_AREA_ID);
        },

        _insertCursorPlaceholder = function() {
            if (!_isComposeAreaInFocus()) {
                return null;
            }

            let range,
                cursorPlaceholder;

            try {
                range = window.getSelection().getRangeAt(0);
            } catch (err) {
                return null;
            }

            cursorPlaceholder = _getPlaceholderSpan();
            range.deleteContents();
            range.insertNode(cursorPlaceholder);

            return cursorPlaceholder;
        },

        _getCursorOffsets = function() {
            let cursorPlaceholder = _insertCursorPlaceholder(),
                cursorOffsets = {};

            if (!cursorPlaceholder) {
                return null;
            }
            
            // MI-24576: traverse all offsets to get absolute offset
            let absoluteOffsetTop = 0;
            let currentElement = cursorPlaceholder;
            do {
                absoluteOffsetTop += currentElement.offsetTop || 0;
                currentElement = currentElement.offsetParent;
            } while (currentElement);

            cursorOffsets.top = absoluteOffsetTop - CURSOR_PADDING;
            cursorOffsets.bottom = cursorOffsets.top + cursorPlaceholder.offsetHeight + CURSOR_PADDING * 2;
            _removeCursorPlaceholder();

            return cursorOffsets;
        },

        _onMessageInput = function(e) {
            // MI-16714: prevent clearing of text when using ios toolbar format options
            if (e.inputType.startsWith("format")) {
                // Save draft on style change
                window.location = "mailsdk:contentDidChange";
                return;
            }

            scrollIfNeeded();
        },

        _onMessageKeyup = function(e) {
            // Prevent clearing of text when cmd+a shortcut, shift, or shift+arrow is used
            if (e.key === "Meta" || e.key === "Shift" || (e.shiftKey && e.key.indexOf("Arrow") === 0)) {
                return;
            }

            scrollIfNeeded();
        },

        _addEventListeners = function() {
            _composeArea.addEventListener("input", _onMessageInput);
            _composeArea.addEventListener("keyup", _onMessageKeyup);
        },

        /// Replace "file:" paths with wk-friendly "localfile:" paths
        _loadLocalFilesIfNeeded = function() {
            let divNodes,
                imgNodes;

            divNodes = document.querySelectorAll("div.inline-attachment-container.attachment-image[data-image-source^='file:']");
            imgNodes = document.querySelectorAll("img.custom-inline-image[src^='file:']");
            divNodes.forEach(function(node) { _replaceAttributeValue(node, "data-image-source", "file:", "localfile:"); });
            imgNodes.forEach(function(node) { _replaceAttributeValue(node, "src", "file:", "localfile:"); });
        },

        _replaceAttributeValue = function(node, attribute, toReplace, replacement) {
            let attributeValue = node.getAttribute(attribute);
            node.setAttribute(attribute, attributeValue.replace(toReplace, replacement));
        },

        _onDOMContentLoaded = function(event) {
            _loadLocalFilesIfNeeded();
            _initElements();
            _addEventListeners();
            // Set completion flag on window
            window.wkMsgComposeCompleted = true;
            
            _postNativeMsg("dom:contentLoaded");
        },
    
        _doScrollIfNeeded = function() {
            let trueViewPortHeight = window.YMUTrueViewportHeight;

            if (!trueViewPortHeight) {
                return;
            }

            let cursorOffsets = _getCursorOffsets();

            if (cursorOffsets) {
                _postNativeMsg("window:scrollToCursor", {
                    trueViewPortHeight: trueViewPortHeight,
                    cursorOffsetTop: cursorOffsets.top,
                    cursorOffsetBottom: cursorOffsets.bottom
                });
            }
        },

        /* PUBLIC METHODS */

        // Scrolling natively (el.scrollTo()) isn't 100% reliable, so have the app do it
        scrollIfNeeded = function() {
            if (_scrollIfNeededTimeout) {
                clearTimeout(_scrollIfNeededTimeout);
            }

            _scrollIfNeededTimeout = setTimeout(_doScrollIfNeeded, SCROLL_DEBOUNCE_TIME);
            
            return true;
        },
        // For smart replies, this method positions the cursor at the end of the smart reply
        setCursorAtEndOfLine = function() {
            var selection = window.getSelection(),
                range = selection.getRangeAt(0);

            range.setStart(_composeArea, 1);
            range.collapse();
            selection.removeAllRanges();
            selection.addRange(range);
            
            return true;
        },
        // Sets current compose content as the initial content of the message
        setCurrentContentAsInitialContent = function() {
            _postNativeMsg("window:initialContent", {
                content: _composeArea.innerHTML
            });
            
            return true;
        },
        // Sets visibility of the drag and drop target area
        setDragAndDropAreaVisibility = function(isVisible) {
            document.body.classList[isVisible ? "add" : "remove"](DND_ACTIVE_CLASS_NAME);
            return true;
        };

    // When html and scripts are loaded
    document.addEventListener("DOMContentLoaded", _onDOMContentLoaded);

    // MARK: - Public methods
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    return {
        scrollIfNeeded: scrollIfNeeded,
        setCursorAtEndOfLine: setCursorAtEndOfLine,
        setCurrentContentAsInitialContent: setCurrentContentAsInitialContent,
        setDragAndDropAreaVisibility: setDragAndDropAreaVisibility
    };
}());
