var YRSEEmailAddressValidator = (function() {
    //  This regex is made global so that it is only evaluated once.
    const EMAIL_REGEX = /^(\\.|[A-Za-z0-9!#%&`_=\/$'*+?^{}|~.-])+$/; // eslint-disable-line no-useless-escape
    const EMAIL_QUOTE_REGEX = /^"(\\"|[^"])+"$/;
    const EMAIL_DOMAINCHARS_REGEX = /^[A-Za-z0-9\-]+$/; // eslint-disable-line no-useless-escape

    /**
     * The format of email addresses is local-part@domain-part.
     * It destructures the email address into local and domain parts
     * @method destructEmailAddressParts
     * @param {string} formatted email address
     * @returns {Object} Object contains local part and domain part
     */
    function destructEmailAddressParts(emailAddress) {
        const atIdx = emailAddress.lastIndexOf('@');
        const localPart = emailAddress.substr(0, atIdx);
        const domainPart = emailAddress.substr(atIdx + 1);

        return { localPart, domainPart };
    }

    /**
     * Validates local part in the email address
     * @method validateLocalPart
     * @param {string} local part of email address
     * @returns {Boolean} truthy if local part is valid
     */
    function validateLocalPart(localPart) {
        // RFC 2822, 3.2.4, 4.4: The local part may include valid characters or escaped (with \) characters
        // RFC 2822, 3.2.5: The local part may consist of a quoted string,
        // i.e. anything within quotes ("), including spaces
        if (!EMAIL_REGEX.test(localPart) && !EMAIL_QUOTE_REGEX.test(localPart)) {
            return false;
        }

        return true;
    }

    /**
     * Validates domain part in the email address
     * @method validateDomainPart
     * @param {string} domain part of email address
     * @returns {Boolean} truthy if domain part is valid
     */
    function validateDomainPart(domainPart) {
        // RFC 1035, 2.3.1: A domain consists of labels separated by dots
        const domainParts = domainPart.split('.');
        const domainLength = domainParts.length;

        if (domainLength < 2) {
            return false;
        }

        // RFC 1035, 2.3.1: Domain labels start with an alphabetic character followed by 0 or more
        // alphanumeric characters or the hyphen (-), ending with an alphanumeric character.
        // NOTE: This allows numeric and hyphen characters at the beginning and end.  This is because I'm
        // too lazy to make the complete check, and also so it will pass 'sfrancis@66.94.238.147'.
        for (let i = 0; i < domainLength; i++) {
            if (!EMAIL_DOMAINCHARS_REGEX.test(domainParts[i])) {
                return false;
            }
        }

        return true;
    }
    /**
     * Validates an email address
     * @method isValidEmail
     * @param {string} email address to validate
     * @returns {Boolean} truthy if email address is valid
     * @remark See http://www.linuxjournal.com/article/9585 for explanation
     */
    function isValidEmail(emailAddress) {
        if (!emailAddress) {
            return false;
        }

        const { localPart, domainPart } = destructEmailAddressParts(emailAddress);

        if (!validateLocalPart(localPart)) {
            return false;
        }

        if (!validateDomainPart(domainPart)) {
            return false;
        }

        return true;
    }
    
    // MARK: - Public methods
    // All methods below must return a value to avoid `Fatal error: Bug in WebKit: Received neither result or failure` when evaluated via `WKWebView.evaluateJavaScript()`
    return {
        isValidEmail: isValidEmail
    };
}());
