const TextSizeAdjustPolyfill = (function() {


/*
    [MI-24811][iPad] Text scaling -webkit-text-size-adjust CSS property is not supported on iPad.
    This is a polyfill style workaround with equivalent functionality.

    How it works:
       1. Find all text elements and [1] record fontSize and optionally lineHeight as custom attributes
         [2] add a class name that will scale the font [see zoomTextContent(), markElementsNeedingFontZoom()]

       2. zoom text by creating a style id="ym-zoomed-text", with a zoom rule for each class name created
          in step 1. [see zoomTextContent(), createTextZoomStyle()]

       3. If the zoom level changes, drop the ym-zoomed-text style, and repeat step 2.

       4. On document window size change, the zoom process needs to be reset and repeated. This is
          because on document size change @media queries may kick in and change the original font sizes.
          [see updateZoomedText()]
 */


    const FONT_SIZE_ATTRIBUTE = 'ym-font-size';
    const LINE_HEIGHT_ATTRIBUTE = 'ym-line-height';
    const ZOOMED_TEXT_ID = 'ym-zoomed-text';

    let zoomRootNode =  null;
    let currentZoomLevel = 100;
    let debounceTimer = null;

    function isNumber(value) {
        return typeof value === 'number' && isFinite(value);
    }

    function hasValue(value) {
        return !(value === null || value === undefined);
    }

    function roundedFloat(value) {
        return parseInt(value) == value ? value : value.toFixed(1);
    }

    function zoomedTextClassNameFor(element) {
        const fontSize = element.getAttribute(FONT_SIZE_ATTRIBUTE) || '';
        const lineHeight = element.getAttribute(LINE_HEIGHT_ATTRIBUTE) || '';

        let className = `${ZOOMED_TEXT_ID}-${String(fontSize).replaceAll('.', '_')}`;
        if (isNumber(parseFloat(lineHeight))) {
            className += `-${String(lineHeight).replaceAll('.', '_')}`;
        }
        return className;
    }

     function markElementsNeedingFontZoom(rootNode) {

        const visitedElementSet = new Set();
        const treeWalker = document.createTreeWalker(rootNode, NodeFilter.SHOW_TEXT, null, false);

        let textNode;
        while (textNode = treeWalker.nextNode()) {
            if (!/\S/.test(textNode.nodeValue)) continue; // empty text node

            const element = textNode.parentNode;
            if (visitedElementSet.has(element)) continue;
            visitedElementSet.add(element);

            const elementComputedStyle = getComputedStyle(element);
            if (elementComputedStyle.display == 'none') continue;

            const fontSizeValue = elementComputedStyle.fontSize || '';
            if (!fontSizeValue.endsWith('px')) continue;

            const fontSize = parseFloat(fontSizeValue);
            if (!isNumber(fontSize)) continue;

            const lineHeightValue = elementComputedStyle.lineHeight || '';
            const lineHeight = lineHeightValue.endsWith('px') ? parseFloat(lineHeightValue) : null;

            element.setAttribute(FONT_SIZE_ATTRIBUTE, `${roundedFloat(fontSize)}`);
            lineHeight && element.setAttribute(LINE_HEIGHT_ATTRIBUTE, `${roundedFloat(lineHeight)}`);

            const className = zoomedTextClassNameFor(element);
            element.classList.add(className);
        }
    }

    function createTextZoomStyle(rootNode, zoomLevel) {

        removeTextZoomStyle();
        if (zoomLevel == 100) return;

        const fontScale = zoomLevel / 100;
        const textElements = document.querySelectorAll(`[${FONT_SIZE_ATTRIBUTE}]`);

        const fontSizeClassNameSet = new Set();
        let cssRules = [];

        // Collect all font size / line height combos
        for (const element of textElements) {
            let fontSize = parseFloat(element.getAttribute(FONT_SIZE_ATTRIBUTE));
            if (!isNumber(fontSize)) continue;

            const className = zoomedTextClassNameFor(element);
            if (fontSizeClassNameSet.has(className)) continue;
            fontSizeClassNameSet.add(className);

            let cssRule = `.${className} { font-size: ${roundedFloat(fontSize * fontScale)}px !important;`;
            const lineHeight = parseFloat(element.getAttribute(LINE_HEIGHT_ATTRIBUTE));
            if (isNumber(lineHeight)) {
                cssRule += ` line-height: ${roundedFloat(lineHeight * fontScale)}px !important;`
            }
            cssRule += ` }`;
            cssRules.push(cssRule);
        }

        // Create a style element with scaled text rules
        let zoomStyle = document.createElement('style');
        zoomStyle.id = ZOOMED_TEXT_ID;
        zoomStyle.setAttribute("zoom", `${zoomLevel}`);
        zoomStyle.textContent = cssRules.join('\n');
        document.head.appendChild(zoomStyle);
    }

    function removeTextZoomStyle() {
        const styleElement = document.getElementById(ZOOMED_TEXT_ID);
        styleElement && styleElement.remove();
    }

    function resetTextZoom() {
        removeTextZoomStyle();

        const textZoomedElements = document.querySelectorAll(`[${FONT_SIZE_ATTRIBUTE}]`);
        for (const element of textZoomedElements) {
            const className = zoomedTextClassNameFor(element);
            element.classList.remove(className);
            element.removeAttribute(FONT_SIZE_ATTRIBUTE);
            element.removeAttribute(LINE_HEIGHT_ATTRIBUTE);
        }

        zoomRootNode = null;
        currentZoomLevel = 100;
    }

    function isTextZoomed() {
        return hasValue(zoomRootNode) && currentZoomLevel != 100;
    }

    // To be called to set or change text zoom scale
    function zoomTextContent({rootNode, zoomLevel}) {

        if (!zoomRootNode) {
            zoomRootNode = rootNode;
            currentZoomLevel = zoomLevel;
            markElementsNeedingFontZoom(rootNode);
            createTextZoomStyle(rootNode, zoomLevel);
            return;
        }

        if (currentZoomLevel != zoomLevel) {
            currentZoomLevel = zoomLevel;
            createTextZoomStyle(rootNode, zoomLevel);
        }
    }

    // To be called on window resize as font size can change from CSS @media rules
    function updateZoomedText({debounce_ms = 0, completion = () => {}}) {
        if (!isTextZoomed()) {
            setTimeout(() => { completion(false) }, 0);
            return;
        }

        const rootNode = zoomRootNode;
        const zoomLevel = currentZoomLevel;

        clearTimeout(debounceTimer);
        debounceTimer = setTimeout( () => {
            resetTextZoom();
            zoomTextContent({rootNode: rootNode, zoomLevel: zoomLevel});
            completion(true);
        }, debounce_ms);
    }

    // Public
    return {
        zoomTextContent: zoomTextContent,
        updateZoomedText: updateZoomedText,
    };
}());
